// ==================================================================
// Implementation of Apple Metal Rendering of the IFS data
// ==================================================================

@import simd;
@import MetalKit;

#import "MetalRenderer.h"
#import "MetalMTKView.h"
#import "MetalShaderTypes.h"

#import "ifs.h"


// The one significant global variable
extern IFS *ifs;


@implementation MetalRenderer
{
  id<MTLDevice> _device;
  id<MTLRenderPipelineState> _pipelineState;
  id<MTLCommandQueue> _commandQueue;
  vector_uint2 _viewportSize;
  id<MTLBuffer> _vertexBuffer;

  //Uniform
  id<MTLBuffer> mvpUniform;
    
  // The number of vertices in our vertex buffer;
  NSUInteger _numVertices;

  float translation_x;
  float translation_y;

  float zoom;

  float rotation_vertical;
  float rotation_horizontal;
}


- (nonnull instancetype)initWithMetalKitView:(nonnull MetalMTKView *)mtkView
{
    self = [super init];
    if(self)
    {
      _device = mtkView.device;
      [self loadMetal:mtkView];
      [mtkView setRenderer:self];
      // initialize camera
      translation_x = 0;
      translation_y = 0;
      zoom = 1;
      rotation_vertical = 0;
      rotation_horizontal = 0;
    }
    return self;
}


+ (nonnull NSData *)generateVertexData
{
  NSUInteger dataSize = ifs->cube_count * sizeof(cube) + ifs->point_count * sizeof(MetalVertex);
  NSMutableData *vertexData = [[NSMutableData alloc] initWithLength:dataSize];

  MetalVertex* current;
  // copy the cube data
  current = vertexData.mutableBytes;
  memcpy(current, ifs->cubeData, sizeof(cube)*ifs->cube_count);
  
  // copy the point data
  current = vertexData.mutableBytes + ifs->cube_count * sizeof(cube);
  memcpy(current, ifs->pointData, sizeof(MetalVertex)*ifs->point_count);
  
  return vertexData;
}


- (void)loadMetal:(nonnull MTKView *)mtkView
{
    mtkView.colorPixelFormat = MTLPixelFormatBGRA8Unorm_sRGB;

    id<MTLLibrary> defaultLibrary = [_device newDefaultLibrary];

    id<MTLFunction> vertexFunction = [defaultLibrary newFunctionWithName:@"vertexShader"];
    id<MTLFunction> fragmentFunction = [defaultLibrary newFunctionWithName:@"fragmentShader"];

    MTLRenderPipelineDescriptor *pipelineStateDescriptor = [[MTLRenderPipelineDescriptor alloc] init];
    pipelineStateDescriptor.label = @"Simple Pipeline";
    pipelineStateDescriptor.vertexFunction = vertexFunction;
    pipelineStateDescriptor.fragmentFunction = fragmentFunction;
    pipelineStateDescriptor.depthAttachmentPixelFormat = MTLPixelFormatDepth32Float;
    // depth buffer still broken...
    //pipelineStateDescriptor.depthStencilPixelFormat = Depth32Float_Stencil8;
    pipelineStateDescriptor.colorAttachments[0].pixelFormat = mtkView.colorPixelFormat;

    NSError *error = NULL;
    _pipelineState = [_device newRenderPipelineStateWithDescriptor:pipelineStateDescriptor error:&error];
    if (!_pipelineState) { NSLog(@"Failed to created pipeline state, error %@", error); }

    [self reGenerate];
  
    _commandQueue = [_device newCommandQueue];
}



- (void)cameraTranslate :(float)tx :(float)ty
{
  translation_x+=tx;
  translation_y+=ty;
}

- (void)cameraRotate :(float)tx :(float)ty
{
  rotation_vertical+=0.5*ty;
  if (rotation_vertical > 89.99) { rotation_vertical = 89.99; }
  if (rotation_vertical < -89.99) { rotation_vertical = -89.99; }
  rotation_horizontal+=0.5*tx;
}

- (void)cameraZoom :(float)tx :(float)ty
{
  zoom *= pow(1.003,ty);
}


- (void)reGenerate
{
  NSData *vertexData = [MetalRenderer generateVertexData];
  _vertexBuffer = [_device newBufferWithLength:vertexData.length
                                       options:MTLResourceStorageModeShared];
  memcpy(_vertexBuffer.contents, vertexData.bytes, vertexData.length);
  _numVertices = vertexData.length / sizeof(MetalVertex);
}


- (void)mtkView:(nonnull MTKView *)view drawableSizeWillChange:(CGSize)size
{
  _viewportSize.x = size.width;
  _viewportSize.y = size.height;
}


- (matrix_float4x4) makeXRotationMatrix:(float) angle {
  matrix_float4x4 m=matrix_identity_float4x4;
  m.columns[1][1] = cos(angle);    m.columns[2][1] = -sin(angle);
  m.columns[1][2] = sin(angle);    m.columns[2][2] = cos(angle);
  return m;
}
- (matrix_float4x4) makeYRotationMatrix:(float) angle {
  matrix_float4x4 m=matrix_identity_float4x4;
  m.columns[0][0] = cos(angle);    m.columns[2][0] = -sin(angle);
  m.columns[0][2] = sin(angle);    m.columns[2][2] = cos(angle);
  return m;
}
- (matrix_float4x4) makeZRotationMatrix:(float) angle {
  matrix_float4x4 m=matrix_identity_float4x4;
  m.columns[0][0] = cos(angle);    m.columns[1][0] = sin(angle);
  m.columns[0][1] = -sin(angle);   m.columns[1][1] = cos(angle);
  return m;
}


- (matrix_float4x4) makeTranslationMatrix:(vector_float3) translation {
  matrix_float4x4 m=matrix_identity_float4x4;
  m.columns[3][0] = translation.x;
  m.columns[3][1] = translation.y;
  m.columns[3][2] = translation.z;
  return m;
}

- (matrix_float4x4) makeScaleMatrix:(float) scale {
  matrix_float4x4 m=matrix_identity_float4x4;
  m.columns[0][0] = scale;
  m.columns[1][1] = scale;
  m.columns[2][2] = scale;
  return m;
}

- (matrix_float4x4) makeOrthographicMatrix:(float) field_of_view :(float) aspect :(float)near :(float)far {
  matrix_float4x4 m=matrix_identity_float4x4;

  float left   = -0.7;
  float right  =  0.7;
  float top    =  0.7;
  float bottom = -0.7;

  m.columns[0][0] =  2.0 / (float)(right-left);
  m.columns[1][1] =  2.0 / (float)(top-bottom);
  m.columns[2][2] =  2.0 / (float)(far-near);

  m.columns[3][0] = -(float)(right+left) / (float)(right-left);
  m.columns[3][1] = -(float)(top+bottom) / (float)(top-bottom);
  m.columns[3][2] = (float)(far+near)   / (float)(far-near);
  m.columns[3][3] = 1;
  
  return m;
}


- (matrix_float4x4) makePerspectiveMatrix:(float) field_of_view :(float) aspect :(float)near :(float)far {
  matrix_float4x4 m=matrix_identity_float4x4;

  // work in progress
  
  float scale = 1.0 / tan(field_of_view * 0.5 * M_PI / 180.0);
  scale = 1;
  
  m.columns[0][0] = scale;
  m.columns[1][1] = scale;

  //m.columns[2][2] = 0;
  m.columns[2][2] = (far) / (far-near);
  //m.columns[2][3] = 0; //0.001;

  m.columns[3][2] = (far*near) / (far-near);
  m.columns[3][3] = 1;
  return m;
}



/// Called whenever the view needs to render a frame
- (void)drawInMTKView:(nonnull MTKView *)view
{

  // Create a new command buffer for each render pass to the current drawable
  id<MTLCommandBuffer> commandBuffer = [_commandQueue commandBuffer];
  commandBuffer.label = @"MyCommand";

  // Obtain a renderPassDescriptor generated from the view's drawable textures
  MTLRenderPassDescriptor *renderPassDescriptor = view.currentRenderPassDescriptor;

  if(renderPassDescriptor != nil)  {

      // set background color to white
      renderPassDescriptor.colorAttachments[0].clearColor = MTLClearColorMake(1.0,1.0,1.0,1.0);
      
      // Create a render command encoder so we can render into something
      id<MTLRenderCommandEncoder> renderEncoder =
        [commandBuffer renderCommandEncoderWithDescriptor:renderPassDescriptor];
      renderEncoder.label = @"MyRenderEncoder";

      [renderEncoder setViewport:(MTLViewport){0.0, 0.0, _viewportSize.x, _viewportSize.y, -1.0, 1.0 }];
      [renderEncoder setRenderPipelineState:_pipelineState];
      [renderEncoder setCullMode:MTLCullModeFront];

      // center the cube
      vector_float3 first_translation = { -0.5, -0.5, -0.5 };

      vector_float3 translation = {-0.01*translation_x,-0.01*translation_y,0};
      float fov = 45;
      float aspect = 1;
      float near = -10;
      float far = 100;
          
      matrix_float4x4 modelMatrix= [self makeTranslationMatrix:first_translation];
          
      matrix_float4x4 worldMatrix=
        matrix_multiply ( [self makeXRotationMatrix:rotation_vertical*M_PI/180],
                          matrix_multiply ( [self makeYRotationMatrix:rotation_horizontal*M_PI/180],
                                            matrix_multiply ( [self makeScaleMatrix:zoom],
                                                              [self makeTranslationMatrix:translation] )));
        
      matrix_float4x4 viewMatrix=matrix_identity_float4x4;
        
      matrix_float4x4 projectiveMatrix=[self makeOrthographicMatrix:fov :aspect :near :far];
 
      matrix_float4x4 modelViewProjectionTransformation
        =matrix_multiply(projectiveMatrix,
                         matrix_multiply(viewMatrix,
                                         matrix_multiply(worldMatrix, modelMatrix)));

      //Load the MVP transformation into the MTLBuffer
      mvpUniform=[_device newBufferWithBytes:(void*)&modelViewProjectionTransformation
                                      length:sizeof(modelViewProjectionTransformation)
                                     options:MTLResourceOptionCPUCacheModeDefault];
        

      // Send our data to the Metal vertex shader
      [renderEncoder setVertexBuffer:_vertexBuffer
                              offset:0
                             atIndex:MetalVertexInputIndexVertices];
      [renderEncoder setVertexBytes:&_viewportSize
                             length:sizeof(_viewportSize)
                            atIndex:MetalVertexInputIndexViewportSize];
      [renderEncoder setVertexBuffer:mvpUniform offset:0 atIndex:2];
        
      int numCubeVertices = 6 * 2 * 3 * ifs->cube_count;
      int numPointVertices = _numVertices - numCubeVertices;
        
      // Draw the cubes
      [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle
                        vertexStart:0
                        vertexCount:numCubeVertices];

      // Draw the points
      [renderEncoder drawPrimitives:MTLPrimitiveTypePoint
                        vertexStart:numCubeVertices
                        vertexCount:numPointVertices];

      [renderEncoder endEncoding];
      [commandBuffer presentDrawable:view.currentDrawable];
    }

  // Finalize rendering here & push the command buffer to the GPU
  [commandBuffer commit];
}

@end
