// ==================================================================
// OpenGL Rendering of the IFS data
// ==================================================================

#include "OpenGLRenderer.h"
#include "OpenGLCamera.h"
#include "OpenGLCanvas.h"

// ====================================================================
OpenGLRenderer::OpenGLRenderer(IFS *_ifs, ArgParser *args) {

  ifs = _ifs;

  OpenGLCanvas::initialize(args,ifs);

  // Initialize the IFS
  setupVBOs();
  
  glClearColor(1.0f, 1.0f, 1.0f, 0.0f);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS); 
  glEnable(GL_CULL_FACE);

  // Create and compile our GLSL program from the shaders
  GLuint programID = LoadShaders( args->path+"/OpenGL.vertexshader", args->path+"/OpenGL.fragmentshader" );
  
  // Get a handle for our "MVP" uniform
  GLuint MatrixID = glGetUniformLocation(programID, "MVP");

  while (!glfwWindowShouldClose(OpenGLCanvas::window))  {
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(programID);

    OpenGLCanvas::camera->glPlaceCamera();

    // Build the matrix to position the camera based on keyboard and mouse input
    glm::mat4 ProjectionMatrix = OpenGLCanvas::camera->getProjectionMatrix();
    glm::mat4 ViewMatrix = OpenGLCanvas::camera->getViewMatrix();
    glm::mat4 ModelMatrix = glm::mat4(1.0);
    glm::mat4 MVP = ProjectionMatrix * ViewMatrix * ModelMatrix;
    
    // pass the matrix to the draw routines (for further editing)
    drawVBOs(MatrixID,MVP);

    // Swap buffers
    glfwSwapBuffers(OpenGLCanvas::window);
    glfwPollEvents();  
  }
  
  cleanupVBOs();
  glDeleteProgram(programID);
  
  // Close OpenGL window and terminate GLFW
  glfwDestroyWindow(OpenGLCanvas::window);
  glfwTerminate();
  exit(EXIT_SUCCESS);
}

// ====================================================================

void OpenGLRenderer::setupVBOs() {
  HandleGLError("enter setupVBOs");
  if (ifs->cubes) {
    setupCube();
  } else {
    setupPoints();
  }
  HandleGLError("leaving setupVBOs");
}

void OpenGLRenderer::drawVBOs(GLuint MatrixID,const glm::mat4 &m) {
  HandleGLError("enter drawVBOs");
  if (ifs->cubes) {
    glUniformMatrix4fv(MatrixID, 1, GL_FALSE, &m[0][0]);
    drawCube();
  } else {
    glUniformMatrix4fv(MatrixID, 1, GL_FALSE, &m[0][0]);
    drawPoints();
  }
  HandleGLError("leaving drawVBOs");
}

void OpenGLRenderer::cleanupVBOs() {
  HandleGLError("enter cleanupVBOs");
  if (ifs->cubes) {
    cleanupCube();
  } else {
    cleanupPoints();
  }
  HandleGLError("leaving cleanupVBOs");
}

// ====================================================================

void OpenGLRenderer::setupPoints() {
  // create a pointer for the VBO
  glGenVertexArrays(1, &VaoId);
  glBindVertexArray(VaoId);
  glGenBuffers(1, &VboId);
  glBindBuffer(GL_ARRAY_BUFFER,VboId); 
  glBufferData(GL_ARRAY_BUFFER,sizeof(VertexPosColor) * ifs->point_count, ifs->pointData,GL_STATIC_DRAW); 
  glEnableVertexAttribArray(0);
  glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(VertexPosColor), 0);
  glEnableVertexAttribArray(1);
  glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, sizeof(VertexPosColor), (GLvoid*)sizeof(glm::vec4));
  // increase the size of the points slightly
  glPointSize(2.0);
  HandleGLError("leaving setupPoints");
}

void OpenGLRenderer::drawPoints() const {
  HandleGLError("enter drawPoints");
  glDrawArrays(GL_POINTS, 0, ifs->point_count);
  HandleGLError("leaving drawPoints");
}

void OpenGLRenderer::cleanupPoints() {
  glDeleteBuffers(1, &VaoId);
  glDeleteBuffers(1, &VboId);
}

// ====================================================================

void OpenGLRenderer::setupCube() {
  HandleGLError("enter setupCube");
  glGenVertexArrays(1, &VaoId);
  glBindVertexArray(VaoId);
  glGenBuffers(1, &VboId);
  glBindBuffer(GL_ARRAY_BUFFER, VboId);
  int sizeOfVertices = sizeof(VertexPosColor) * ifs->cube_count * 12 * 3;
  glBufferData(GL_ARRAY_BUFFER, sizeOfVertices, ifs->cubeData, GL_STATIC_DRAW);
  glEnableVertexAttribArray(0);
  glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(VertexPosColor), 0);
  glEnableVertexAttribArray(1);
  glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, sizeof(VertexPosColor), (GLvoid*)sizeof(glm::vec4));
  HandleGLError("leaving setupCube");
}

void OpenGLRenderer::drawCube() const {
  HandleGLError("in drawCube");
  glDrawArrays(GL_TRIANGLES, 0, NUM_CUBE_VERTS * ifs->cube_count);
  HandleGLError("leaving drawCube");
}

void OpenGLRenderer::cleanupCube() {
  glDeleteBuffers(1, &VaoId);
  glDeleteBuffers(1, &VboId);
}

// ====================================================================
